/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Dorgem - Webcam Capture application                                    //
//  Copyright (C) 2001-2007 Frank Fesevur                                  //
//                                                                         //
//  This program is free software; you can redistribute it and/or modify   //
//  it under the terms of the GNU General Public License as published by   //
//  the Free Software Foundation; either version 2 of the License, or      //
//  (at your option) any later version.                                    //
//                                                                         //
//  This program is distributed in the hope that it will be useful,        //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of         //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the           //
//  GNU General Public License for more details.                           //
//                                                                         //
//  You should have received a copy of the GNU General Public License      //
//  along with this program; if not, write to the Free Software            //
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#include "Dorgem.h"

#ifdef __WXMSW__
#include <wx/msw/private.h>
#include <shlobj.h>
#endif

#include "../TinyXml/tinyxml.h"
#include "XmlEntries.h"

#include "DorgemDlg.h"
#include "ProfileManager.h"

IMPLEMENT_APP(CDorgemApp);

BEGIN_EVENT_TABLE(CDorgemApp, wxApp)
  //EVT_TIMER(TIMER_ID, CDorgemApp::OnTimer)
END_EVENT_TABLE()

/////////////////////////////////////////////////////////////////////////////
//

CDorgemApp::CDorgemApp()
{
  m_pChecker = NULL;
  m_pProfileMngr = NULL;
}

CDorgemApp::~CDorgemApp()
{
  if (m_pChecker != NULL)
    delete m_pChecker;

  if (m_pProfileMngr != NULL)
  {
    m_pProfileMngr->RemoveAll();
    delete m_pProfileMngr;
  }
}

/////////////////////////////////////////////////////////////////////////////
//

bool CDorgemApp::OnInit()
{
  // First check if another instance is already running
  m_pChecker = new wxSingleInstanceChecker(wxT(PRODUCT_NAME) + wxGetUserId());
  if (m_pChecker->IsAnotherRunning())
    return(false);

#ifdef __WXMSW__
  // Get the full path to the application path
  wxChar szPath[_MAX_PATH];
  DWORD dwSize = _MAX_PATH;
  if (GetModuleFileName(wxGetInstance(), szPath, dwSize) == 0)
    return(false);
  m_strAppPath = szPath;
  int iIndex = m_strAppPath.Find(wxT('\\'), TRUE);
  wxASSERT(iIndex != -1);
  m_strAppPath = m_strAppPath.Left(iIndex + 1);
#endif

  GetXmlSettingsPath();

  // Initialize the locale support
  m_locale.Init(wxLANGUAGE_DEFAULT);
  m_locale.AddCatalogLookupPathPrefix(m_strAppPath + wxT("locale"));
  m_locale.AddCatalog(wxT(PRODUCT_NAME));

  // Initialize the wxImage handler
  ::wxInitAllImageHandlers();

  // Read the settings from the XML file
  if (!ReadSettingsFromXml())
    return(false);

  // Read the profile settings
  m_pProfileMngr = new CProfileManager;
  wxASSERT(m_pProfileMngr != NULL);
  if (!m_pProfileMngr->ReadEvents())
    return(false);

  // Create the window
  CDorgemDlg *pDlg = new CDorgemDlg(NULL);

  // and show it (the frames, unlike simple controls, are not shown when
  // created initially)
  pDlg->Show(true);

  // Write the profile settings
  m_pProfileMngr->WriteEvents();

  // success: wxApp::OnRun() will be called which will enter the main message
  // loop and the application will run. If we returned FALSE here, the
  // application would exit immediately.
  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

int CDorgemApp::OnExit()
{
  WriteSettingsToXml();
  return(0);
}

/////////////////////////////////////////////////////////////////////////////
//

void CDorgemApp::GetXmlSettingsPath()
{
  m_fnSettingsPath.Clear();

#ifdef __WXMSW__
  wxChar buffer[MAX_PATH + 2];

  if (SUCCEEDED(SHGetFolderPath(0, CSIDL_APPDATA, 0, SHGFP_TYPE_CURRENT, buffer)))
  {
    m_fnSettingsPath = wxFileName(buffer, wxT(""));
    m_fnSettingsPath.AppendDir(wxT(PRODUCT_NAME));
  }
#endif

  if (!m_fnSettingsPath.IsOk())
  {
    // This does not work, so do it the wxString way
    //m_fnSettingsPath = wxFileName::GetHomeDir();
    //m_fnSettingsPath.AppendDir(wxString(wxT(".")) + wxT(PRODUCT_NAME));

    wxString str = wxFileName::GetHomeDir();
    str += wxFileName::GetPathSeparator();
    str += wxT(".");
    str += wxT(PRODUCT_NAME);
    str += wxFileName::GetPathSeparator();
    m_fnSettingsPath = str;
  }

  if (!m_fnSettingsPath.DirExists())
    m_fnSettingsPath.Mkdir();
}

/////////////////////////////////////////////////////////////////////////////
// Helper functions to make it more easy to get stuff from the XML

int CDorgemApp::GetAttributeInt(TiXmlElement* pElem, wxString strAttr, int iDefault)
{
  wxString strTmp = pElem->Attribute(strAttr);
  long lVal;
  bool bValid = strTmp.ToLong(&lVal);
  return(bValid ? (int) lVal : iDefault);
}

bool CDorgemApp::GetAttributeBool(TiXmlElement* pElem, wxString strAttr, bool bDefault)
{
  return(GetAttributeInt(pElem, strAttr, bDefault ? 1 : 0) != 0);
}

wxString CDorgemApp::GetAttributeString(TiXmlElement* pElem, wxString strAttr, wxString strDefault)
{
  wxString strTmp = pElem->Attribute(strAttr);
  return(strTmp.IsEmpty() ? strDefault : strTmp);
}

wxColour CDorgemApp::GetAttributeColour(TiXmlElement* pElem, wxString strAttr, wxColour colDefault)
{
  wxString strTmp = pElem->Attribute(strAttr);
  return(strTmp.IsEmpty() ? colDefault : wxColour(strTmp));
}

/////////////////////////////////////////////////////////////////////////////
//

bool CDorgemApp::ReadSettingsFromXml()
{
  wxFileName path = m_fnSettingsPath;
  path.SetFullName(wxT("Settings.xml"));

  TiXmlDocument doc(path.GetFullPath());
  if (!doc.LoadFile())
  {
    // If file not found, the initial defaults must be used
    if (doc.ErrorId() == 2)
      return(true);
    else
    {
      wxLogDebug(wxT("Could not load file 'Settings.xml'. Error='%s'. Exiting.\n"), doc.ErrorDesc());
      return(false);
    }
  }

  // Get the "Settings" element.
  TiXmlNode* nodeSettings = doc.FirstChild(g_strXmlSettings);
  if (nodeSettings == NULL)
    return(false);

  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CDorgemApp::WriteSettingsToXml()
{
  wxFileName path = m_fnSettingsPath;
  path.SetFullName(wxT("Settings.xml"));

  TiXmlDocument doc(path.GetFullPath());
  TiXmlElement elemSettings(g_strXmlSettings);
  elemSettings.SetAttribute(g_strXmlVersion, SETTING_XML);

  doc.InsertEndChild(elemSettings);
  return(doc.SaveFile());
}

////////////////////////////////////////////////////////////////////////////////
// A *very* simple crypt routine.
// Only used to prevent plain text passwords in XML files

void CDorgemApp::Crypt(wxString& strMessage)
{
  wxString strKey = wxT(PRODUCT_NAME);

  int iLenK = strKey.Length(), iLenM = strMessage.Length(), i, j, d;

  for (i = 0, j = 0; i < iLenM; i++)
  {
    d = strMessage[i] + strKey[j] - 64;
    d -= (d / 93) * 93;
    strMessage.SetChar(i, wxChar(d + 32));
    j++;
    j -= (j / iLenK) * iLenK;
  }
}

////////////////////////////////////////////////////////////////////////////////
// DeCrypt a text crypted by the Crypt() above

void CDorgemApp::DeCrypt(wxString& strMessage)
{
  wxString strKey = wxT(PRODUCT_NAME);

  int iLenK = strKey.Length(), iLenM = strMessage.Length(), i, j, d;

  for (i = 0, j = 0; i < iLenM; i++)
  {
    d = strMessage[i] - strKey[j];
    if (d < 0)
      d += 93;
    strMessage.SetChar(i, wxChar(d + 32));
    j++;
    j -= (j / iLenK) * iLenK;
  }
}

////////////////////////////////////////////////////////////////////////////////
//

wxString CDorgemApp::DateFormat(wxString str)
{
  wxDateTime tmNow = wxDateTime::Now();
  //tmNow.SetLanguage(m_wLang);

  // Will ruin a %%P, but not a real problem at the moment
  str.Replace("%P", wxT(PRODUCT_NAME));
  str.Replace("%V", wxT(PRODUCT_VERSION_STR));

  return(tmNow.Format(str));
}
