/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Dorgem - Webcam Capture application                                    //
//  Copyright (C) 2001-2007 Frank Fesevur                                  //
//                                                                         //
//  This program is free software; you can redistribute it and/or modify   //
//  it under the terms of the GNU General Public License as published by   //
//  the Free Software Foundation; either version 2 of the License, or      //
//  (at your option) any later version.                                    //
//                                                                         //
//  This program is distributed in the hope that it will be useful,        //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of         //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the           //
//  GNU General Public License for more details.                           //
//                                                                         //
//  You should have received a copy of the GNU General Public License      //
//  along with this program; if not, write to the Free Software            //
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
// StoreEvent.cpp : implementation file

#include "Dorgem.h"
#include "StoreEvent.h"
#include "CaptionManager.h"
#include "CaptionBitmap.h"
#include "CaptionText.h"

#define DEFAULT_INTERVAL       30
#define DEFAULT_JPEG_QUALITY   70
#define DEFAULT_MD_SENSITIVITY  3

/////////////////////////////////////////////////////////////////////////////
// Initialize the members

CStoreEvent::CStoreEvent() : CDorgemEvent()
{
  m_iType = STORE_NONE;
  m_bBusy = false;
  m_tmLastUpdate = wxDateTime::Now();
  m_bResetTime = true;
  m_tInterval = DEFAULT_INTERVAL;
  m_iLogLevel = 0;
  m_iHistory = 0;
  m_iQuality = DEFAULT_JPEG_QUALITY;
  m_bMotionDetection = false;
  m_iMDSensitivity = DEFAULT_MD_SENSITIVITY;
  m_pCaptions = NULL;
}

/////////////////////////////////////////////////////////////////////////////
//

CStoreEvent::~CStoreEvent()
{
/*
  // Close the log-file
  if (m_fileLog.m_hFile != CFile::hFileNull)
    m_fileLog.Close();
*/

  // Save the captions
  if (m_pCaptions != NULL)
    delete m_pCaptions;
}

/////////////////////////////////////////////////////////////////////////////
// Does this event need to be stored?

bool CStoreEvent::NeedStore()
{
  if (!m_bEnabled)
    return(false);

  wxDateTime tmNow = wxDateTime::Now();
  return(m_tmLastUpdate.IsEarlierThan(tmNow.Subtract(wxTimeSpan(0, 0, m_tInterval, 0))));
}

/////////////////////////////////////////////////////////////////////////////
// Reset the timer, just like a Store() has occurred

void CStoreEvent::ResetTimer()
{
  if (m_bResetTime)
    m_tmLastUpdate = wxDateTime::Now();
}

/////////////////////////////////////////////////////////////////////////////
// Process very changes to the image. Need to be used *just* before saving.

void CStoreEvent::ProcessImage()
{
  // Do we need to rotate the image?
  if (m_iRotate != 0)
  {
    // Do the 90-180-270 with the Rotate90().
    // This is done for speed reasons
    switch (m_iRotate)
    {
      case  90:
        m_CurImage = m_CurImage.Rotate90();
        break;

      case 180:
        m_CurImage = m_CurImage.Rotate90().Rotate90();
        break;

      case 270:
        m_CurImage = m_CurImage.Rotate90(false);
        break;

      default: // All others with the Rotate()
      {
        wxPoint pt(m_CurImage.GetWidth() / 2, m_CurImage.GetHeight() / 2);
        m_CurImage = m_CurImage.Rotate((double) m_iRotate * (M_PI / 180.0), pt);
      }
    }
  }

  // Put the captions on the bitmap.
  if (m_pCaptions != NULL)
  {
    CCaptionEvent* pEvent = (CCaptionEvent*) m_pCaptions->GetFirstEvent();
    while (pEvent != NULL)
    {
      pEvent->Overlay(m_CurImage);
      pEvent = (CCaptionEvent*) m_pCaptions->GetNextEvent();
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
// Routine to subtract image from the reference image this will give us an
// image with only the changes showing. It's not quite a simple subtraction,
// we look at each color and look for color changes as well.
// Significantly changed pixels are set to bright white. Result is a b/w
// image with additional blob filtering to look for areas of change
// (marked as red pixels). Each blob is a pixel surrounded by 8 neighbors
//
// An adapted version of the motion detection of 'gspy' a gnome security
// camera application (http://gspy.sourceforge.net), also a GPL application
//
// The function assumes the bitmap is 24 bits RGB
// If the image is larger then 358x288, you are advised to downsize it

bool CStoreEvent::MotionDetection()
{
  // Do we use motion detection?
  if (!m_bMotionDetection)
    return(true);

  // Create a copy of the captured image
  wxASSERT(m_CurImage.IsOk());
  wxImage imgCurrent = m_CurImage;

  // Do we need to put an overlay on the captured image?
  if (m_fnMDOverlay.IsOk())
  {
    // Load the bitmap
    wxImage imgOverlay;
    if (imgOverlay.LoadFile(m_fnMDOverlay.GetFullPath()))
      imgCurrent.Paste(imgOverlay, 0, 0);
  }

  // The very first time, the previous image is not set yet.
  // Obviously then we have motion detected.
  if (!m_PrevImage.IsOk())
  {
    // Save the current image as previous
    m_PrevImage = imgCurrent;
    return(true);
  }

  // Get the information about the two images
  int iHeight = imgCurrent.GetHeight();
  int iWidth = imgCurrent.GetWidth();
  unsigned char* pSrc = imgCurrent.GetData();
  unsigned char* pRef = m_PrevImage.GetData();

  // Create a result buffer
  int iSize = iHeight * iWidth * 3;
  unsigned char* pResBits = new unsigned char[iSize];
  unsigned char* pRes = pResBits;

  int x, y, red, green, blue;
  unsigned char cResult;
  float fResult = 0.0f;
  for (x = 0; x < iWidth; x++)
  {
    for (y = 0; y < iHeight; y++)
    {
      // give the cpu something to do...
      // compute rms diff between all pixels
      // detection of color or intensity changes detected
      red = (*pRef > *pSrc ? *pRef - *pSrc : *pSrc - *pRef);    // red
      pRef++; pSrc++;
      green = (*pRef > *pSrc ? *pRef - *pSrc : *pSrc - *pRef);  // green
      pRef++; pSrc++;
      blue  = (*pRef > *pSrc ? *pRef - *pSrc : *pSrc - *pRef);  // blue
      pRef++; pSrc++;
      cResult = (unsigned char) sqrt((double) (red * red + green * green + blue * blue) / 3);
      *pRes++ = cResult;     // make b/w image
      *pRes++ = cResult;
      *pRes++ = cResult;
      fResult += (float) cResult;
    }
  }
  // calculate normalized difference over picture
  fResult = fResult / (float) (iHeight * iWidth);

  // detect and mark pixels that are significantly different
  unsigned char cThreshold = 25;
  int iPixCount = 0;
  pSrc = pResBits;
  while (pSrc < pResBits + iSize)
  {
    if (*pSrc > cThreshold)
    {
      iPixCount++;
      *pSrc = 255;        // alarm active areas in white
      *(pSrc + 1) = 255;
      *(pSrc + 2) = 255;
    }
    pSrc += 3;          // only need to look at reds of each pixel
                        // because g and b are set to the same thing
  }

  int iAlarmZoneX1 = iWidth;            // init bounding box of alarm blobs
  int iAlarmZoneX2 = -1;
  int iAlarmZoneY1 = iHeight;
  int iAlarmZoneY2 = -1;

  // To eliminate small changes (wind blown objects) from causing so much
  // havoc, we will try to eliminate false alarms by scanning for blobs
  // A blob is a 3x3 pixel area, all must be 'hot' (white)
  int i, iBlobCount = 0;
  pSrc = pResBits + iWidth * 3;
  while (pSrc < (pResBits + iSize - iWidth * 3))
  {
    for (i = 0; i < 1; i++)
    {
      if (*pSrc < 255) break;                       // central pixel must be lite
      if (*(pSrc - iWidth * 3) < 255) break;        // pixel above must be lite
      if (*(pSrc + iWidth * 3) < 255) break;        // pixel below must be lite
      if (*(pSrc - 3) < 255) break;                 // pixel to left must be lite
      if (*(pSrc + 3) < 255) break;                 // pixel to right must be lite
      if (*(pSrc - iWidth * 3 + 3) < 255) break;    // pixel to NE must be lite
      if (*(pSrc - iWidth * 3 - 3) < 255) break;    // pixel to NW must be lite
      if (*(pSrc + iWidth * 3 + 3) < 255) break;    // pixel to SE must be lite
      if (*(pSrc + iWidth * 3 - 3) < 255) break;    // pixel to SW must be lite

      // mark blob as blob on screen
      *(pSrc + 2) = 0;  // shut off blue
      *(pSrc + 1) = 0;  // shut off green
      iBlobCount++;
      x = ((pSrc - pResBits) % (iWidth * 3)) / 3;
      y = (pSrc - pResBits) / (iWidth * 3);
      if (x < iAlarmZoneX1)        // set first occurance location
        iAlarmZoneX1 = x;
      if (iAlarmZoneX2 < x)        // keep highest occurance
        iAlarmZoneX2 = x;
      if (y < iAlarmZoneY1)        // set first occurance location
        iAlarmZoneY1 = y;
      if (iAlarmZoneY2 < y)        // keep highest occurance
        iAlarmZoneY2 = y;
    }
    pSrc += 3;
  }

  delete [] pResBits;

  // Are there enough differences detected?
  if (iBlobCount > m_iMDSensitivity)
  {
    // Save the current image (incl overlay) as previous
    m_PrevImage = imgCurrent;
    return(true);
  }

  return(false);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CStoreEvent::Read(TiXmlElement* pElemProfile)
{
  if (!CDorgemEvent::Read(pElemProfile))
    return(false);

  m_tInterval = wxGetApp().GetAttributeInt(pElemProfile, g_strXmlInterval, DEFAULT_INTERVAL);
  m_bResetTime = wxGetApp().GetAttributeBool(pElemProfile, g_strXmlResetTime, true);
  m_iQuality = wxGetApp().GetAttributeInt(pElemProfile, g_strXmlQuality, DEFAULT_JPEG_QUALITY);
  m_iRotate = wxGetApp().GetAttributeInt(pElemProfile, g_strXmlRotate, 0);

  // Motion detection
  TiXmlNode* pNodeMotion = pElemProfile->FirstChildElement(g_strXmlMotion);
  if (pNodeMotion != NULL)
  {
    TiXmlElement* pElem = pNodeMotion->ToElement();
    wxASSERT(pElem != NULL);
    m_bMotionDetection = wxGetApp().GetAttributeBool(pElem, g_strXmlEnabled, false);
    m_iMDSensitivity = wxGetApp().GetAttributeInt(pElem, g_strXmlSensitivity, DEFAULT_MD_SENSITIVITY);
    m_fnMDOverlay = wxGetApp().GetAttributeString(pElem, g_strXmlFile);
  }

  // Read the captions for this event (if any)
  if (m_pCaptions == NULL)
  {
    m_pCaptions = new CCaptionManager;
    wxASSERT(m_pCaptions != NULL);
  }
  if (!m_pCaptions->ReadEvents(pElemProfile))
  {
    delete m_pCaptions;
    m_pCaptions = NULL;
  }

  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CStoreEvent::Write(TiXmlElement* pElemProfile)
{
  CDorgemEvent::Write(pElemProfile);

  pElemProfile->SetAttribute(g_strXmlInterval, m_tInterval);
  pElemProfile->SetAttribute(g_strXmlResetTime, m_bResetTime);
  pElemProfile->SetAttribute(g_strXmlQuality, m_iQuality);
  pElemProfile->SetAttribute(g_strXmlRotate, m_iRotate);

  // Add the motion detection settings
  TiXmlElement elemMotion(g_strXmlMotion);
  elemMotion.SetAttribute(g_strXmlEnabled, m_bMotionDetection);
  elemMotion.SetAttribute(g_strXmlSensitivity, m_iMDSensitivity);
  elemMotion.SetAttribute(g_strXmlFile, m_fnMDOverlay.GetFullPath());
  pElemProfile->InsertEndChild(elemMotion);

  // Add the captions for this event
  if (m_pCaptions != NULL)
    m_pCaptions->WriteEvents(pElemProfile);

  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

/*
void CStoreEvent::Log(UINT uLevel, LPCSTR szLog)
{
  // Are we in the current log level?
  if (uLevel > m_uLogLevel)
    return;

  // Is the logfile opened?
  if (!OpenLogFile(TRUE))
    return;

  // Construct the time stamp
  CTM tmNow;
  CString str = tmNow.Format("%Y-%0m-%0d %0H:%0M:%0S ");

  // Add the type
  switch (m_iType)
  {
    case STORE_FILE:    str += "File Event";              break;
    case STORE_FTP:     str += "FTP Event";               break;
    case STORE_EMAIL:   str += "EMail Event";             break;
    case STORE_PROGRAM: str += "External Program Event";  break;
    case STORE_HTTP:    str += "HTTP Event";              break;
    default: ASSERT(FALSE);
  }

  // Add the name and text to log
  str += "(" + m_strName + "): " + szLog;

  // Finally do the real logging
  m_fileLog.WriteString(str + "\n");
  m_fileLog.Flush();
}
*/

////////////////////////////////////////////////////////////////////////////
// Open or close the log file

/*
void CStoreEvent::SetLogFile(CString str, BOOL bNew)
{
  if (str.IsEmpty())
  {
    m_fileLog.Close();
    m_strLogFile.Empty();
  }
  else
  {
    // If the path doesn't contain a '\' there is no full path given.
    // So add the app-path to the log file.
    if (str.Find("\\") == -1)
      m_strLogFile = GetApp()->m_strAppPath + str;
    else
      m_strLogFile = str;
  }
}
*/

/////////////////////////////////////////////////////////////////////////////
//

/*
bool CStoreEvent::OpenLogFile(bool bNew)
{
  // Is the logfile name set
  if (m_strLogFile.IsEmpty())
    return(false);

  // Is the file already opened?
  if (m_fileLog.m_hFile != CFile::hFileNull)
    return(true);

  // Create the proper flag
  UINT uFlags = CFile::modeCreate | CFile::modeWrite | CFile::shareDenyNone;
  if (!bNew)
    uFlags |= CFile::modeNoTruncate;

  // Make the filename time related and open the file
  CTM tmNow;
  m_fileLog.Open(tmNow.Format(m_strLogFile), uFlags);

  if (!bNew)
    m_fileLog.SeekToEnd();

  return(true);
}
*/

/////////////////////////////////////////////////////////////////////////////
// Do nothing. This is a place holder. If an event can do a move, it needs
// to implement this function in the derived class.

bool CStoreEvent::MoveFileTo(wxString WXUNUSED(strFrom), wxString WXUNUSED(strTo))
{
  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

void CStoreEvent::MaintainHistory(wxString strFile)
{
  // Do we need to maintain a history?
  if (m_iHistory == 0)
    return;

/*
  // First strip (and keep) the extension
  int iIndex = strFile.ReverseFind('.');
  if (iIndex < 0)
    return;

  CString strFileBase = strFile.Left(iIndex);
  CString strExt = strFile.Mid(iIndex);
  CString strNum, strNumPrev, strFrom, strTo;
  strNumPrev.Format("%d", m_uHistory);

  // Try to rename the old versions
  for (int i = m_uHistory - 1; i >= 0; i--)
  {
    if (i == 0)
      strNum.Empty();
    else
      strNum.Format("%d", i);

    strFrom = strFileBase;
    strFrom += strNum;
    strFrom += strExt;

    strTo = strFileBase;
    strTo += strNumPrev;
    strTo += strExt;

    MoveFileTo(strFrom, strTo);

    strNumPrev = strNum;
  }
*/
}

/////////////////////////////////////////////////////////////////////////////
//

/*
void CStoreEvent::GetInterval(time_t& tInterval, int& iBetween)
{
  // First try to find out if the interval is a whole hour or minute
  CTM tm;
  time_t tBetween;
  for (int i = BETWEEN_HOURS; i < BETWEEN_SECS; i++)
  {
    tBetween = tm.GetBetweenOffset(i);
    if (m_tInterval % tBetween == 0)
    {
      tInterval = m_tInterval / tBetween;
      iBetween = i;
      return;
    }
  }

  // return the seconds
  tInterval = m_tInterval;
  iBetween = BETWEEN_SECS;
}
*/

/////////////////////////////////////////////////////////////////////////////
//

/*
void CStoreEvent::SetInterval(int iInterval, int iBetween)
{
  CTM tm;
  m_tInterval = iInterval;
  m_tInterval *= tm.GetBetweenOffset(iBetween);
}
*/

/////////////////////////////////////////////////////////////////////////////
//

void CStoreEvent::DeleteCaptions()
{
  if (!HasCaptions())
    return;

  delete m_pCaptions;
  m_pCaptions = NULL;
}
