/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Dorgem - Webcam Capture application                                    //
//  Copyright (C) 2001-2007 Frank Fesevur                                  //
//                                                                         //
//  This program is free software; you can redistribute it and/or modify   //
//  it under the terms of the GNU General Public License as published by   //
//  the Free Software Foundation; either version 2 of the License, or      //
//  (at your option) any later version.                                    //
//                                                                         //
//  This program is distributed in the hope that it will be useful,        //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of         //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the           //
//  GNU General Public License for more details.                           //
//                                                                         //
//  You should have received a copy of the GNU General Public License      //
//  along with this program; if not, write to the Free Software            //
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
// WebServer.cpp : implementation file

#include "Dorgem.h"
#include "WebServer.h"
#include "Version.h"
#include "VideoWindow.h"

enum
{
  SERVER_ID = 200,
  SOCKET_ID
};

BEGIN_EVENT_TABLE(CWebServer, wxEvtHandler)
  EVT_SOCKET(SERVER_ID, CWebServer::OnServerEvent)
  EVT_SOCKET(SOCKET_ID, CWebServer::OnSocketEvent)
END_EVENT_TABLE()

#define MAXBUF 1024
#define CRLF   "\x0d\x0a"

/////////////////////////////////////////////////////////////////////////////
// Initialize the members

CWebServer::CWebServer(wxWindow* pWnd, CVideoWindow* pVidWnd) : CStoreEvent()
{
  wxASSERT(pWnd != NULL);
  wxASSERT(pVidWnd != NULL);

  m_iPort = 8080;
  m_bIsRunning = false;
  m_pServer = NULL;
  m_pWnd = pWnd;
  m_vidCapWin = pVidWnd;
  m_iConnections = 0;
}

CWebServer::~CWebServer()
{
  Stop();
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::Start()
{
  wxIPV4address addr;

  addr.AnyAddress();
  addr.Service(m_iPort);

  m_pServer = new wxSocketServer(addr);
  if (!m_pServer->IsOk())
  {
    delete m_pServer;
    m_pServer = NULL;
    m_bIsRunning = false;
    wxLogDebug("Staring Web Server failed");
    return(false);
  }

  m_pWnd->PushEventHandler(this);

  m_pServer->SetEventHandler(*this, SERVER_ID);
  m_pServer->SetNotify(wxSOCKET_CONNECTION_FLAG);
  m_pServer->Notify(true);
  wxLogDebug("Web Server started, listening on port %d", m_iPort);

  m_bIsRunning = true;
  return(m_bIsRunning);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::Stop()
{
  if (!m_bIsRunning)
    return(false);

  m_pWnd->PopEventHandler();
  m_pServer->Destroy();
  delete m_pServer;
  m_pServer = NULL;
  m_bIsRunning = false;
  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

void CWebServer::OnServerEvent(wxSocketEvent& WXUNUSED(event))
{
  // Accept the new connection and get the socket pointer
  wxSocketBase* pSock = m_pServer->Accept(false);

  // Tell the new socket how and where to process its events
  if (pSock != NULL)
  {
    pSock->SetEventHandler(*this, SOCKET_ID);
    pSock->SetNotify(wxSOCKET_INPUT_FLAG | wxSOCKET_LOST_FLAG);
    pSock->Notify(true);
    m_iConnections++;
  }
}

/////////////////////////////////////////////////////////////////////////////
//

void CWebServer::OnSocketEvent(wxSocketEvent& event)
{
  wxSocketBase* pSock = event.GetSocket();

  // Process the event
  switch(event.GetSocketEvent())
  {
    case wxSOCKET_INPUT:
    {
      // Read the data
      char buf[MAXBUF];
      memset(buf, 0, MAXBUF);
      pSock->Read(buf, sizeof(buf));

      // Try to parse the HTTP request
      if (ParseRequest(buf, MAXBUF))
      {
        // If the camera is connected get a image from it,
        // else get an 'off-line' image
        wxImage img;
        if (m_vidCapWin->IsDeviceConnected())
          m_vidCapWin->SnapshotTowxImage(img);
        else
          img.LoadFile("C:/Mijn documenten/Admea Logo.bmp");

        // Write that image
        SendImage(pSock, &img, true);
      }
      else
        SendStatus(pSock, 501, wxT("Not Implemented"));

      // We are done with the socket, destroy it
      pSock->Destroy();
      m_iConnections--;
      break;
    }
    case wxSOCKET_LOST:
    {
      pSock->Destroy();
      m_iConnections--;
      break;
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::ParseRequest(const char* pBuf, size_t iSize)
{
  // Just in case
  if (strlen(pBuf) > iSize)
    return(false);

  wxArrayString strLines;
  wxString strLine;

  while (*pBuf)
  {
    if (*pBuf != '\r')
    {
      if (*pBuf == '\n')
      {
        if (!strLine.IsEmpty())
        {
          strLines.Add(strLine);
          strLine.Empty();
        }
      }
      else
        strLine += *pBuf;
    }
    pBuf++;
  }

  if (strLines.GetCount() == 0)
    return(false);

  if (strLines[0].Left(3) != wxT("GET"))
    return(false);

  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::SendImage(wxSocketBase* pSock, wxImage* pImg, bool bStillImage)
{
  // Store the image as a JPEG in a memory stream
  wxMemoryOutputStream outStream;
  pImg->SetOption(wxT("quality"), m_iQuality);
  pImg->SaveFile(outStream, wxBITMAP_TYPE_JPEG);

  // Create an input memory stream based upon this stream
  wxMemoryInputStream inStream(outStream);

  //
  if (bStillImage)
  {
    SendStatus(pSock, 200, "OK");
    SendHeader(pSock, "Content-Type", "image/jpeg");
    SendHeader(pSock, "Content-Length", inStream.GetSize());
    SendHeader(pSock, "Last-Modified", GetHttpDate());
    //SendHeader(pSock, "Cache-Control", "no-cache");         // HTTP 1.1
    //SendHeader(pSock, "Cache-Control", "private");          // HTTP 1.1
    //SendHeader(pSock, "Cache-Control", "must-revalidate");  // HTTP 1.1
    //SendHeader(pSock, "Pragma", "no-cache");                // HTTP 1.0
    //SendHeader(pSock, "Expires", 0);
    //if (m_uStillRefreshRate != 0)
    //  SendHeader(pSock, "Refresh", m_uStillRefreshRate);
    SendString(pSock, CRLF);
  }
  else
  {
    SendHeader(pSock, "Content-type", "multipart/x-mixed-replace;boundary=DorgemRandomString");

    //SendHeader(pSock, "Cache-Control", "no-cache");         // HTTP 1.1
    //SendHeader(pSock, "Cache-Control", "private");          // HTTP 1.1
    //SendHeader(pSock, "Cache-Control", "must-revalidate");  // HTTP 1.1
    //SendHeader(pSock, "Pragma", "no-cache");                // HTTP 1.0
    //SendString(pSock, CRLF);

    // For streaming, send the extra headers
    SendString(pSock, "--DorgemRandomString" + wxString(CRLF));
    SendHeader(pSock, "Content-type", "image/jpeg");
    SendString(pSock, CRLF);
  }

  // Go through this stream and write the data to the socket
  size_t iReadSize;
  char buf[MAXBUF];
  while (!inStream.Eof())
  {
    inStream.Read(buf, MAXBUF);
    iReadSize = inStream.LastRead();
    if (iReadSize > 0)
    {
      pSock->Write(buf, iReadSize);
      if (pSock->Error())
        return(false);
    }
  }

  return(true);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::SendString(wxSocketBase* pSock, wxString strMsg)
{
  pSock->Write(strMsg.GetData(), strMsg.Length());
  return(!pSock->Error());
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::SendStatus(wxSocketBase* pSock, int iStatus, wxString strMsg)
{
  wxString strTmp;
  strTmp.Printf("HTTP/1.0 %d %s%s", iStatus, strMsg, CRLF);
  if (!SendString(pSock, strTmp))
    return(false);

  // Send the server name
  strTmp = wxT(PRODUCT_NAME);
  strTmp += "/";
  strTmp += wxT(PRODUCT_VERSION_STR);
  strTmp.Replace(" ", "-");  // replace all spaces with dashes
  if (!SendHeader(pSock, "Server", strTmp))
    return(false);

  // Send the date
  return(SendHeader(pSock, "Date", GetHttpDate()));
}

/////////////////////////////////////////////////////////////////////////////
//

wxString CWebServer::GetHttpDate()
{
  int iCurLang = wxLocale::GetSystemLanguage();
  wxLocale locale(wxLANGUAGE_ENGLISH);

  wxDateTime tmNow = wxDateTime::Now();
  wxString strTmp = tmNow.Format("%a, %d %b %Y %H:%M:%S GMT", wxDateTime::UTC);

  wxLocale locale2(iCurLang);
  return(strTmp);
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::SendHeader(wxSocketBase* pSock, wxString strName, wxString strValue)
{
  wxString strTmp = strName;
  strTmp += ": ";
  strTmp += strValue;
  strTmp += CRLF;
  return(SendString(pSock, strTmp));
}

/////////////////////////////////////////////////////////////////////////////
//

bool CWebServer::SendHeader(wxSocketBase* pSock, wxString strName, int iValue)
{
  return(SendString(pSock, wxString::Format("%s: %d%s", strName, iValue, CRLF)));
}
